/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.hayabusa.html;

import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.hayabusa.db.DBTableModel;
import org.opengion.hayabusa.db.DBColumn;
import org.opengion.fukurou.util.StringUtil;
import org.opengion.fukurou.util.Attributes;
import org.opengion.fukurou.model.Formatter;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays ;

/**
 * ViewMarker インターフェース の実装オブジェクトです。
 * これを,共通のスーパークラスとして 各種表示フォーム(例：HTML表示等)に使います。
 *
 * このクラスは、setter/getterメソッドのデフォルト実装を提供しています。
 * 各種表示フォームに対応したサブクラス上で, create() をオーバーライドして下さい。
 *
 * @og.group 画面表示
 *
 * @version  4.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public class ViewMarker_MARKER implements ViewMarker {

	private List<Attributes>		markData	= null;		// 4.0.0 (2005/08/31)
//	private Map<Integer,int[]>		formMap_c	= new HashMap<Integer,int[]>();	// 3.5.6.1 (2004/06/25)
//	private Map<Integer,String[]>	formMap_f	= new HashMap<Integer,String[]>();	// 3.5.6.1 (2004/06/25)
	private Map<Integer,Formatter>	formMap		= new HashMap<Integer,Formatter>();
	private DBTableModel		table       = null;
	private int[]				markCmlNo	= null;
	private int[]				isMark		= null;
	private static final int	MARK_NULL   = -1;	// マーカー未設定
	private static final int	MARK_TRUE   = 1;	// マーカー作成
	private static final int	MARK_FALSE  = 0;	// マーカー作成せず
	// 3.5.2.0 (2003/10/20)
	private String[]			markKey		= null;
	private String[]			markLists	= null;
	private String[]			instrVals	= null;		// 3.8.8.1 (2007/01/06)
	private int[]				markListNo	= null;
	private boolean[]			useFmtDeco	= null;		// 5.6.3.0 (2013/04/01) [$XXXX],[#XXXX]機能を有効にするかどうか(true:有効)

	private Map<Integer,List<Integer>>	clmMap	= new HashMap<Integer,List<Integer>>();	// 4.0.0 (2005/08/31)

	/**
	 * 内容をクリア(初期化)します。
	 *
	 * @og.rev 3.1.1.0 (2003/03/28) 同期メソッド(synchronized付き)を非同期に変更する。
	 * @og.rev 3.5.2.0 (2003/10/20) markLists,markListNo,markKey属性を追加
	 * @og.rev 3.5.6.1 (2004/06/25) formMap属性を追加
	 * @og.rev 3.8.8.1 (2007/01/06) instrVals属性を追加
	 * @og.rev 5.6.3.0 (2013/04/01) useFmtDeco属性を追加
	 *
	 */
	public void clear() {
		markData	= null;		// 4.0.0 (2005/08/31)
//		formMap_c	= new HashMap<Integer,int[]>();	// 3.5.6.1 (2004/06/25)
//		formMap_f	= new HashMap<Integer,String[]>();	// 3.5.6.1 (2004/06/25)
		formMap		= new HashMap<Integer,Formatter>();
		table		= null;
		isMark		= null;
		markKey		= null;
		markLists	= null;
		instrVals	= null;		// 3.8.8.1 (2007/01/06)
		markListNo	= null;
		clmMap		= new HashMap<Integer,List<Integer>>();	// 4.0.0 (2005/08/31)
		useFmtDeco	= null;		// 5.6.3.0 (2013/04/01) [$XXXX],[#XXXX]機能を有効にするかどうか(true:有効)
	}

	/**
	 * カラムに対するマーカーアトリビュートをセットします。
	 *
	 * @og.rev 3.1.0.0 (2003/03/20) Hashtable を使用している箇所で、非同期でも構わない箇所を、HashMap に置換え。
	 * @og.rev 3.1.1.0 (2003/03/28) 同期メソッド(synchronized付き)を非同期に変更する。
	 *
	 * @param	attri	アトリビュート
	 */
	public void addAttribute( final Attributes attri ) {
		if( markData == null ) { markData = new ArrayList<Attributes>(); }
		markData.add( attri );
	}

	/**
	 * 内部に DBTableModel をセットします。
	 *
	 * @og.rev 3.1.1.0 (2003/03/28) 同期メソッド(synchronized付き)を非同期に変更する。
	 * @og.rev 3.5.2.0 (2003/10/20) markLists,markListNo,markKey属性を追加
	 * @og.rev 3.5.6.1 (2004/06/25) DBTableModel の再設定に対応。
	 * @og.rev 3.8.8.1 (2007/01/06) instrVals属性を追加
	 * @og.rev 5.6.3.0 (2013/04/01) useFmtDeco属性を追加
	 *
	 * @param  tbl DBTableModelオブジェクト
	 */
	public void setDBTableModel( final DBTableModel tbl ) {
		table = tbl;
		int count = markData.size();			// 4.0.0 (2005/08/31)

		isMark		= new int[ count ];
		markKey		= new String[ count ];
		markCmlNo	= new int[ count ];
		markLists	= new String[ count ];
		instrVals	= new String[ count ];
		markListNo	= new int[ count ];
		useFmtDeco	= new boolean[ count ];		// 5.6.3.0 (2013/04/01) [$XXXX],[#XXXX]機能を有効にするかどうか(true:有効)

		Arrays.fill( isMark,MARK_FALSE );	// マーカーの表示可否
		Arrays.fill( markCmlNo,-1 );		// マーカーの可否を判断するカラム番号
		Arrays.fill( useFmtDeco,false );	// [$XXXX],[#XXXX]機能を無効にする。(互換性のため)

		for( int intKey=0; intKey<count; intKey++ ) {
			Attributes attri = markData.get( intKey );

			String column = attri.get( "column" );
			int clm = table.getColumnNo( column );
			List<Integer> list = clmMap.get( clm );
			if( list == null ) { list = new ArrayList<Integer>(); }
			list.add( intKey );
			clmMap.put( clm,list );

			String body = attri.get( "body" );
//			makeFormat( intKey,body );
			Formatter formatter = new Formatter( table );
			formatter.setFormat( body );
			formMap.put( intKey, formatter );

			makeOnMarkFormat( intKey,attri );

			useFmtDeco[intKey] = "true".equalsIgnoreCase( attri.get( "useFormatDeco" ) );	// 5.6.3.0 (2013/04/01)
		}
	}

	/**
	 * 指定の行列に対するマーカー文字列を返します。
	 * この値は,すでにマーカー文字列処理されている為, RendererValue で
	 * 変換する必要はありません。
	 * 引数の value はそのカラムの値として利用されます。この値は、修飾済みの
	 * 値を与えることが可能です。
	 *
	 * @og.rev 3.5.6.1 (2004/06/25) formMap属性を使用します。
	 * @og.rev 3.8.8.1 (2007/01/06) instrVals属性を追加
	 * @og.rev 5.3.9.0 (2011/09/01) カラム名の先頭に'$'を付加した場合に、URLEncodeされた値を返すように対応
	 * @og.rev 5.6.3.0 (2013/04/01) useFmtDeco属性を追加([$XXXX],[#XXXX]機能を有効にするかどうか)
	 *
	 * @param   row 指定の行
	 * @param   clm 指定の列
	 * @param   value カラムの値
	 *
	 * @return  row行，colum列 のマーカー文字列
	 */
	public String getMarkerString( final int row,final int clm,final String value ) {
		int intKey = isOnMark(row,clm) ;
		if( intKey < 0 ) { return value; }

		Formatter formatter = formMap.get( intKey );
		int[]    clmNo  = formatter.getClmNos();
		String[] format = formatter.getFormat();

		char[] types = formatter.getType();

//		int[] clmNo = formMap_c.get( intKey );
//		String[] format = formMap_f.get( intKey );

		StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_LARGE );
		int j=0;
		String val ;
		for( ; j<clmNo.length; j++ ) {
			if( clm == clmNo[j] ) {
				val = value;
			}
			else {
//				val = table.getValue(row,clmNo[j]);
				val = formatter.getValue(row,clmNo[j]);
			}

			// 5.6.3.0 (2013/04/01) useFmtDeco属性を追加(trueの場合は、[$XXXX],[#XXXX]機能を有効にする)
			if( useFmtDeco[intKey] ) {
				DBColumn dbClm = table.getDBColumn( clmNo[j] );
				if( types[j] == '$' ) {
					val = dbClm.getRendererValue( row,val );
				}
				else if( types[j] == '#' ) {
					val = dbClm.getLabel();
				}
			}
			// false が以前と同じ処理（互換処理）ただし、view などのフォーマット処理とは異なる。
			else {
				// 5.3.9.0 (2011/09/01) カラム名の先頭に'$'を付加した場合URLEncodeされた値を返すように対応
				if( types[j] == '$' ) {
					val = StringUtil.urlEncode( val );
				}
			}

			buf.append( format[j] );
			buf.append( val );
		}
		if( j < format.length ) { buf.append( format[j] ); }
		String rtn = StringUtil.replace( buf.toString(),"{I}",String.valueOf( row ) );

		// 3.8.8.1 (2007/01/06) instrVals属性を追加
		if( instrVals[intKey] != null ) {
			String[] vals = StringUtil.csv2Array( instrVals[intKey],' ' );
			for( int i=0; i<vals.length; i++ ) {
				String css = "<span class=\"instr" + i + "\">" + vals[i] + "</span>";
				rtn = StringUtil.replace( rtn,vals[i],css );
			}
		}
		return rtn ;
	}

	/**
	 * マーカーフォーマットを作成します。
	 *
	 * @og.rev 3.5.6.1 (2004/06/25) formMap属性を使用します。
	 * @og.rev 3.8.8.1 (2007/01/06) countTokensの方法を見直します。
	 *
	 * @param	intKey	カラムキーのカラム番号
	 * @param	fmt	マーカーフォーマット
	 */
//	private void makeFormat( final int intKey,final String fmt ) {
//		int start = 0;
//		int index = fmt.indexOf( '[' );
//		List<String> clmNoList  = new ArrayList<String>();
//		List<String> formatList = new ArrayList<String>();
//		while( index >= 0 ) {
//			int end = fmt.indexOf( ']',index );
//			if( end < 0 ) {
//				String errMsg = "[ と ] との対応関係がずれています。" + HybsSystem.CR
//							+ "format=[" + fmt + "] : index=" + index ;
//				throw new HybsSystemException( errMsg );
//			}
//
//			// [ より前方の文字列は、formatList へ
//			if( index > 0 ) { formatList.add( fmt.substring( start,index ) ); }
//			else 			{ formatList.add( "" );							  }
//
//			// [XXXX] の XXXX部分を処理
//			clmNoList.add( fmt.substring( index+1,end ) );
//
//			start = end+1 ;
//			index = fmt.indexOf( '[',start );
//		}
//		// ] の後方部分は、formatList へ
//		formatList.add( fmt.substring( start ) );
//
//		String[] format = formatList.toArray( new String[formatList.size()] );
//		String[] clmNm  = clmNoList.toArray( new String[clmNoList.size()] );
//
//		int size = clmNm.length ;
//		int[] clmNo = new int[ size ];
//		for( int i=0; i<size; i++ ) {
//			clmNo[i]  = table.getColumnNo( clmNm[i] );
//		}
//
//		formMap_c.put( intKey, clmNo );
//		formMap_f.put( intKey, format );
//	}

	/**
	 * マーカーを作成する/作成しないの指定カラム番号を求めます。
	 * また、int[列番号] isMark を初期化します。
	 *
	 * @og.rev 3.5.2.0 (2003/10/20) markLists,markListNo,markKey属性を追加
	 * @og.rev 3.8.8.1 (2007/01/06) instrVals属性を追加
	 *
	 * @param	intKey	カラムキーの番号
	 * @param	attri	アトリビュート
	 */
	private void makeOnMarkFormat( final int intKey,final Attributes attri ) {
		String onMark   = attri.get( "onMark" );
		String markList = attri.get( "markList" );
		instrVals[intKey] = attri.get( "instrVals" );	// 3.8.8.1 (2007/01/06)

		// 3.5.6.0 (2004/06/18) nullポインタの参照外しバグの対応
		// このロジックで値が設定済みであれば、以下の処理は不要である。
		isMark[intKey] = MARK_NULL;
		if( onMark == null || onMark.length() == 0 ||
			markList == null || markList.length() == 0 ) {
				isMark[intKey] = MARK_FALSE;
				return ;	// 3.5.6.0 (2004/06/18)
		}
		else if( onMark.charAt( 0 ) != '[' && markList.charAt( 0 ) != '[' ) {
			isMark[intKey] = ( markList.indexOf( onMark ) >= 0 ) ? MARK_TRUE : MARK_FALSE;
			return ;	// 3.5.6.0 (2004/06/18)
		}

		if( onMark.charAt( 0 ) == '[' ) {
			markCmlNo[intKey] = table.getColumnNo( onMark.substring( 1,onMark.length()-1 ));
		}
		else {
			markCmlNo[intKey]  = -1;
			markKey[intKey]    = onMark ;
		}

		if( markList.charAt( 0 ) == '[' ) {
			markListNo[intKey] = table.getColumnNo( markList.substring( 1,markList.length()-1 ));
		}
		else {
			markListNo[intKey] = -1;
			markLists[intKey] = markList;
		}
	}

	/**
	 * マーカーを作成するかどうかを判断します。
	 * int[列番号] isMark には、 未設定 FALSE TRUE の状態を持っており、
	 * 列でマーカーを作成する状態が固定の場合(例えば,onMark属性がデフォルト "true" の場合)
	 * カラムに関係なく、同じ値を返すときに、使用します。
	 *
	 * @og.rev 3.5.2.0 (2003/10/20) markLists,markListNo,markKey属性を追加
	 * @og.rev 3.5.4.0 (2003/11/25) onMark ,markList が null(またはゼロストリング)の場合は、false とする。
	 * @og.rev 4.0.0.0 (2005/08/31) 同一カラムの複数登録を許可します。
	 *
	 * @param	row	列番号
	 * @param	clm	カラムキーの名称
	 *
	 * @return	処理するリスト番号、-1 の場合は、該当なし
	 */
	private int isOnMark( final int row,final int clm ) {
		List<Integer> list = clmMap.get( clm );
		if( list == null ) { return -1; }

		for( int i=0; i<list.size(); i++ ) {
			int intKey = list.get( i );
			if( isMark[intKey] != MARK_NULL ) {
				if( isMark[intKey] == MARK_TRUE ) { return intKey; }
				else { continue; }
			}

			final String onMark ;
			if( markCmlNo[intKey] < 0 ) { onMark = markKey[intKey] ; }
			else { onMark = table.getValue( row,markCmlNo[intKey] ); }

			// 3.5.4.0 (2003/11/25) 追加
			if( onMark == null || onMark.length() == 0 ) { continue; }

			final String markList ;
			if( markListNo[intKey] < 0 ) { markList = markLists[intKey] ; }
			else { markList = table.getValue( row,markListNo[intKey] ); }

			// 3.5.4.0 (2003/11/25) 修正
			if( markList == null || markList.length() == 0 ) { continue; }

			if( markList.indexOf( onMark ) >= 0 ) { return intKey; }
		}
		return -1;
	}
}
